@@ -3,29 +3,33 @@ package ebpf
 import (
 	"errors"
 	"fmt"
+	"io/ioutil"
 	"math"
 	"os"
 	"path/filepath"
+	"reflect"
 	"sort"
-	"strings"
 	"testing"
 	"unsafe"
 
 	"github.com/cilium/ebpf/internal"
+	"github.com/cilium/ebpf/internal/btf"
 	"github.com/cilium/ebpf/internal/testutils"
 	"github.com/cilium/ebpf/internal/unix"
+
+	qt "github.com/frankban/quicktest"
 )
 
-func TestMain(m *testing.M) {
-	err := unix.Setrlimit(unix.RLIMIT_MEMLOCK, &unix.Rlimit{
-		Cur: unix.RLIM_INFINITY,
-		Max: unix.RLIM_INFINITY,
-	})
-	if err != nil {
-		fmt.Println("WARNING: Failed to adjust rlimit, tests may fail")
+var (
+	spec1 = &MapSpec{
+		Name:       "foo",
+		Type:       Hash,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 1,
+		Pinning:    PinByName,
 	}
-	os.Exit(m.Run())
-}
+)
 
 func TestMap(t *testing.T) {
 	m := createArray(t)
@@ -66,6 +70,197 @@ func TestMap(t *testing.T) {
 	}
 }
 
+func TestBatchAPIArray(t *testing.T) {
+	if err := haveBatchAPI(); err != nil {
+		t.Skipf("batch api not available: %v", err)
+	}
+	m, err := NewMap(&MapSpec{
+		Type:       Array,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 2,
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer m.Close()
+
+	var (
+		nextKey      uint32
+		keys         = []uint32{0, 1}
+		values       = []uint32{42, 4242}
+		lookupKeys   = make([]uint32, 2)
+		lookupValues = make([]uint32, 2)
+		deleteKeys   = make([]uint32, 2)
+		deleteValues = make([]uint32, 2)
+	)
+
+	count, err := m.BatchUpdate(keys, values, nil)
+	if err != nil {
+		t.Fatalf("BatchUpdate: %v", err)
+	}
+	if count != len(keys) {
+		t.Fatalf("BatchUpdate: expected count, %d, to be %d", count, len(keys))
+	}
+
+	var v uint32
+	if err := m.Lookup(uint32(0), &v); err != nil {
+		t.Fatal("Can't lookup 0:", err)
+	}
+	if v != 42 {
+		t.Error("Want value 42, got", v)
+	}
+
+	count, err = m.BatchLookup(nil, &nextKey, lookupKeys, lookupValues, nil)
+	if err != nil {
+		t.Fatalf("BatchLookup: %v", err)
+	}
+	if count != len(lookupKeys) {
+		t.Fatalf("BatchLookup: returned %d results, expected %d", count, len(lookupKeys))
+	}
+	if nextKey != lookupKeys[1] {
+		t.Fatalf("BatchLookup: expected nextKey, %d, to be the same as the lastKey returned, %d", nextKey, lookupKeys[1])
+	}
+	if !reflect.DeepEqual(keys, lookupKeys) {
+		t.Errorf("BatchUpdate and BatchLookup keys disagree: %v %v", keys, lookupKeys)
+	}
+	if !reflect.DeepEqual(values, lookupValues) {
+		t.Errorf("BatchUpdate and BatchLookup values disagree: %v %v", values, lookupValues)
+	}
+
+	_, err = m.BatchLookupAndDelete(nil, &nextKey, deleteKeys, deleteValues, nil)
+	if !errors.Is(err, ErrNotSupported) {
+		t.Fatalf("BatchLookUpDelete: expected error %v, but got %v", ErrNotSupported, err)
+	}
+}
+
+func TestBatchAPIHash(t *testing.T) {
+	if err := haveBatchAPI(); err != nil {
+		t.Skipf("batch api not available: %v", err)
+	}
+	m, err := NewMap(&MapSpec{
+		Type:       Hash,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 10,
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer m.Close()
+
+	var (
+		nextKey      uint32
+		keys         = []uint32{0, 1}
+		values       = []uint32{42, 4242}
+		lookupKeys   = make([]uint32, 2)
+		lookupValues = make([]uint32, 2)
+		deleteKeys   = make([]uint32, 2)
+		deleteValues = make([]uint32, 2)
+	)
+
+	count, err := m.BatchUpdate(keys, values, nil)
+	if err != nil {
+		t.Fatalf("BatchUpdate: %v", err)
+	}
+	if count != len(keys) {
+		t.Fatalf("BatchUpdate: expected count, %d, to be %d", count, len(keys))
+	}
+
+	var v uint32
+	if err := m.Lookup(uint32(0), &v); err != nil {
+		t.Fatal("Can't lookup 0:", err)
+	}
+	if v != 42 {
+		t.Error("Want value 42, got", v)
+	}
+
+	count, err = m.BatchLookup(nil, &nextKey, lookupKeys, lookupValues, nil)
+	if !errors.Is(err, ErrKeyNotExist) {
+		t.Fatalf("BatchLookup: expected %v got %v", ErrKeyNotExist, err)
+	}
+	if count != len(lookupKeys) {
+		t.Fatalf("BatchLookup: returned %d results, expected %d", count, len(lookupKeys))
+	}
+	sort.Slice(lookupKeys, func(i, j int) bool { return lookupKeys[i] < lookupKeys[j] })
+	if !reflect.DeepEqual(keys, lookupKeys) {
+		t.Errorf("BatchUpdate and BatchLookup keys disagree: %v %v", keys, lookupKeys)
+	}
+	sort.Slice(lookupValues, func(i, j int) bool { return lookupValues[i] < lookupValues[j] })
+	if !reflect.DeepEqual(values, lookupValues) {
+		t.Errorf("BatchUpdate and BatchLookup values disagree: %v %v", values, lookupValues)
+	}
+
+	count, err = m.BatchLookupAndDelete(nil, &nextKey, deleteKeys, deleteValues, nil)
+	if !errors.Is(err, ErrKeyNotExist) {
+		t.Fatalf("BatchLookupAndDelete: expected %v got %v", ErrKeyNotExist, err)
+	}
+	if count != len(deleteKeys) {
+		t.Fatalf("BatchLookupAndDelete: returned %d results, expected %d", count, len(deleteKeys))
+	}
+	sort.Slice(deleteKeys, func(i, j int) bool { return deleteKeys[i] < deleteKeys[j] })
+	if !reflect.DeepEqual(keys, deleteKeys) {
+		t.Errorf("BatchUpdate and BatchLookupAndDelete keys disagree: %v %v", keys, deleteKeys)
+	}
+	sort.Slice(deleteValues, func(i, j int) bool { return deleteValues[i] < deleteValues[j] })
+	if !reflect.DeepEqual(values, deleteValues) {
+		t.Errorf("BatchUpdate and BatchLookupAndDelete values disagree: %v %v", values, deleteValues)
+	}
+
+	if err := m.Lookup(uint32(0), &v); !errors.Is(err, ErrKeyNotExist) {
+		t.Fatalf("Lookup should have failed with error, %v, instead error is %v", ErrKeyNotExist, err)
+	}
+}
+
+func TestBatchAPIMapDelete(t *testing.T) {
+	if err := haveBatchAPI(); err != nil {
+		t.Skipf("batch api not available: %v", err)
+	}
+	m, err := NewMap(&MapSpec{
+		Type:       Hash,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 10,
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer m.Close()
+
+	var (
+		keys   = []uint32{0, 1}
+		values = []uint32{42, 4242}
+	)
+
+	count, err := m.BatchUpdate(keys, values, nil)
+	if err != nil {
+		t.Fatalf("BatchUpdate: %v", err)
+	}
+	if count != len(keys) {
+		t.Fatalf("BatchUpdate: expected count, %d, to be %d", count, len(keys))
+	}
+
+	var v uint32
+	if err := m.Lookup(uint32(0), &v); err != nil {
+		t.Fatal("Can't lookup 0:", err)
+	}
+	if v != 42 {
+		t.Error("Want value 42, got", v)
+	}
+
+	count, err = m.BatchDelete(keys, nil)
+	if err != nil {
+		t.Fatalf("BatchDelete: %v", err)
+	}
+	if count != len(keys) {
+		t.Fatalf("BatchDelete: expected %d deletions got %d", len(keys), count)
+	}
+
+	if err := m.Lookup(uint32(0), &v); !errors.Is(err, ErrKeyNotExist) {
+		t.Fatalf("Lookup should have failed with error, %v, instead error is %v", ErrKeyNotExist, err)
+	}
+}
+
 func TestMapClose(t *testing.T) {
 	m := createArray(t)
 
@@ -95,25 +290,26 @@ func TestMapCloneNil(t *testing.T) {
 
 func TestMapPin(t *testing.T) {
 	m := createArray(t)
+	c := qt.New(t)
 	defer m.Close()
 
 	if err := m.Put(uint32(0), uint32(42)); err != nil {
 		t.Fatal("Can't put:", err)
 	}
 
-	tmp := tempBPFFS(t)
-
-	// Issue 51: pad path out to a power of two, to avoid having a
-	// trailing zero at the end of the allocation which holds the string.
-	path := tmp + string(filepath.Separator)
-	path += strings.Repeat("a", 32-len(path))
+	tmp := testutils.TempBPFFS(t)
+	path := filepath.Join(tmp, "map")
 
 	if err := m.Pin(path); err != nil {
 		t.Fatal(err)
 	}
+
+	pinned := m.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+
 	m.Close()
 
-	m, err := LoadPinnedMap(path)
+	m, err := LoadPinnedMap(path, nil)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -128,6 +324,270 @@ func TestMapPin(t *testing.T) {
 	}
 }
 
+func TestNestedMapPin(t *testing.T) {
+	m, err := NewMap(&MapSpec{
+		Type:       ArrayOfMaps,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 2,
+		InnerMap: &MapSpec{
+			Type:       Array,
+			KeySize:    4,
+			ValueSize:  4,
+			MaxEntries: 1,
+		},
+	})
+	testutils.SkipIfNotSupported(t, err)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer m.Close()
+
+	tmp, err := ioutil.TempDir("/sys/fs/bpf", "ebpf-test")
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer os.RemoveAll(tmp)
+
+	path := filepath.Join(tmp, "nested")
+	if err := m.Pin(path); err != nil {
+		t.Fatal(err)
+	}
+	m.Close()
+
+	m, err = LoadPinnedMap(path, nil)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer m.Close()
+}
+
+func TestNestedMapPinNested(t *testing.T) {
+	if _, err := NewMap(&MapSpec{
+		Type:       ArrayOfMaps,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 2,
+		InnerMap: &MapSpec{
+			Name:       "inner",
+			Type:       Array,
+			KeySize:    4,
+			ValueSize:  4,
+			MaxEntries: 1,
+			Pinning:    PinByName,
+		},
+	}); err == nil {
+		t.Error("Inner maps should not be pinnable")
+	}
+}
+
+func TestMapPinMultiple(t *testing.T) {
+	tmp := testutils.TempBPFFS(t)
+	c := qt.New(t)
+
+	spec := spec1.Copy()
+
+	m1, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+	if err != nil {
+		t.Fatal("Can't create map:", err)
+	}
+	defer m1.Close()
+	pinned := m1.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+
+	newPath := filepath.Join(tmp, "bar")
+	err = m1.Pin(newPath)
+	c.Assert(err, qt.IsNil)
+	oldPath := filepath.Join(tmp, spec.Name)
+	if _, err := os.Stat(oldPath); err == nil {
+		t.Fatal("Previous pinned map path still exists:", err)
+	}
+	m2, err := LoadPinnedMap(newPath, nil)
+	c.Assert(err, qt.IsNil)
+	defer m2.Close()
+}
+
+func TestMapPinWithEmptyPath(t *testing.T) {
+	m := createArray(t)
+	c := qt.New(t)
+	defer m.Close()
+
+	err := m.Pin("")
+
+	c.Assert(err, qt.Not(qt.IsNil))
+}
+
+func TestMapPinFailReplace(t *testing.T) {
+	tmp := testutils.TempBPFFS(t)
+	c := qt.New(t)
+	spec := spec1.Copy()
+	spec2 := spec1.Copy()
+	spec2.Name = spec1.Name + "bar"
+
+	m, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+	if err != nil {
+		t.Fatal("Failed to create map:", err)
+	}
+	defer m.Close()
+	m2, err := NewMapWithOptions(spec2, MapOptions{PinPath: tmp})
+	if err != nil {
+		t.Fatal("Failed to create map2:", err)
+	}
+	defer m2.Close()
+	c.Assert(m.IsPinned(), qt.Equals, true)
+	newPath := filepath.Join(tmp, spec2.Name)
+
+	c.Assert(m.Pin(newPath), qt.Not(qt.IsNil), qt.Commentf("Pin didn't"+
+		" fail new path from replacing an existing path"))
+}
+
+func TestMapUnpin(t *testing.T) {
+	tmp := testutils.TempBPFFS(t)
+	c := qt.New(t)
+	spec := spec1.Copy()
+
+	m, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+	if err != nil {
+		t.Fatal("Failed to create map:", err)
+	}
+	defer m.Close()
+
+	pinned := m.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+	path := filepath.Join(tmp, spec.Name)
+	m2, err := LoadPinnedMap(path, nil)
+	c.Assert(err, qt.IsNil)
+	defer m2.Close()
+
+	if err = m.Unpin(); err != nil {
+		t.Fatal("Failed to unpin map:", err)
+	}
+	if _, err := os.Stat(path); err == nil {
+		t.Fatal("Pinned map path still exists after unpinning:", err)
+	}
+}
+
+func TestMapLoadPinned(t *testing.T) {
+	tmp := testutils.TempBPFFS(t)
+	c := qt.New(t)
+
+	spec := spec1.Copy()
+
+	m1, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+	c.Assert(err, qt.IsNil)
+	defer m1.Close()
+	pinned := m1.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+
+	path := filepath.Join(tmp, spec.Name)
+	m2, err := LoadPinnedMap(path, nil)
+	c.Assert(err, qt.IsNil)
+	defer m2.Close()
+	pinned = m2.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+}
+
+func TestMapLoadPinnedUnpin(t *testing.T) {
+	tmp := testutils.TempBPFFS(t)
+	c := qt.New(t)
+
+	spec := spec1.Copy()
+
+	m1, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+	c.Assert(err, qt.IsNil)
+	defer m1.Close()
+	pinned := m1.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+
+	path := filepath.Join(tmp, spec.Name)
+	m2, err := LoadPinnedMap(path, nil)
+	c.Assert(err, qt.IsNil)
+	defer m2.Close()
+	err = m1.Unpin()
+	c.Assert(err, qt.IsNil)
+	err = m2.Unpin()
+	c.Assert(err, qt.IsNil)
+}
+
+func TestMapLoadPinnedWithOptions(t *testing.T) {
+	// Introduced in commit 6e71b04a8224.
+	testutils.SkipOnOldKernel(t, "4.14", "file_flags in BPF_OBJ_GET")
+
+	array := createArray(t)
+	defer array.Close()
+
+	tmp := testutils.TempBPFFS(t)
+
+	path := filepath.Join(tmp, "map")
+	if err := array.Pin(path); err != nil {
+		t.Fatal(err)
+	}
+	if err := array.Put(uint32(0), uint32(123)); err != nil {
+		t.Fatal(err)
+	}
+	array.Close()
+
+	t.Run("read-only", func(t *testing.T) {
+		array, err := LoadPinnedMap(path, &LoadPinOptions{
+			ReadOnly: true,
+		})
+		testutils.SkipIfNotSupported(t, err)
+		if err != nil {
+			t.Fatal("Can't load map:", err)
+		}
+		defer array.Close()
+
+		if err := array.Put(uint32(0), uint32(1)); !errors.Is(err, unix.EPERM) {
+			t.Fatal("Expected EPERM from Put, got", err)
+		}
+	})
+
+	t.Run("write-only", func(t *testing.T) {
+		array, err := LoadPinnedMap(path, &LoadPinOptions{
+			WriteOnly: true,
+		})
+		testutils.SkipIfNotSupported(t, err)
+		if err != nil {
+			t.Fatal("Can't load map:", err)
+		}
+		defer array.Close()
+
+		var value uint32
+		if err := array.Lookup(uint32(0), &value); !errors.Is(err, unix.EPERM) {
+			t.Fatal("Expected EPERM from Lookup, got", err)
+		}
+	})
+}
+
+func TestMapPinFlags(t *testing.T) {
+	tmp := testutils.TempBPFFS(t)
+
+	spec := &MapSpec{
+		Name:       "map",
+		Type:       Array,
+		KeySize:    4,
+		ValueSize:  4,
+		MaxEntries: 1,
+		Pinning:    PinByName,
+	}
+
+	m, err := NewMapWithOptions(spec, MapOptions{
+		PinPath: tmp,
+	})
+	qt.Assert(t, err, qt.IsNil)
+	m.Close()
+
+	_, err = NewMapWithOptions(spec, MapOptions{
+		PinPath: tmp,
+		LoadPinOptions: LoadPinOptions{
+			Flags: math.MaxUint32,
+		},
+	})
+	if !errors.Is(err, unix.EINVAL) {
+		t.Fatal("Invalid flags should trigger EINVAL:", err)
+	}
+}
+
 func createArray(t *testing.T) *Map {
 	t.Helper()
 
@@ -156,23 +616,13 @@ func TestMapQueue(t *testing.T) {
 	}
 	defer m.Close()
 
-	if err := m.Put(nil, uint32(42)); err != nil {
-		t.Fatal("Can't put 42:", err)
-	}
-
-	if err := m.Put(nil, uint32(4242)); err != nil {
-		t.Fatal("Can't put 4242:", err)
+	for _, v := range []uint32{42, 4242} {
+		if err := m.Put(nil, v); err != nil {
+			t.Fatalf("Can't put %d: %s", v, err)
+		}
 	}
 
 	var v uint32
-	if err := m.Lookup(nil, &v); err != nil {
-		t.Fatal("Can't lookup element:", err)
-	}
-	if v != 42 {
-		t.Error("Want value 42, got", v)
-	}
-
-	v = 0
 	if err := m.LookupAndDelete(nil, &v); err != nil {
 		t.Fatal("Can't lookup and delete element:", err)
 	}
@@ -180,16 +630,9 @@ func TestMapQueue(t *testing.T) {
 		t.Error("Want value 42, got", v)
 	}
 
-	if err := m.Lookup(nil, &v); err != nil {
-		t.Fatal("Can't lookup element:", err)
-	}
-	if v != 4242 {
-		t.Error("Want value 4242, got", v)
-	}
-
 	v = 0
-	if err := m.LookupAndDelete(nil, &v); err != nil {
-		t.Fatal("Can't lookup and delete element:", err)
+	if err := m.LookupAndDelete(nil, unsafe.Pointer(&v)); err != nil {
+		t.Fatal("Can't lookup and delete element using unsafe.Pointer:", err)
 	}
 	if v != 4242 {
 		t.Error("Want value 4242, got", v)
@@ -365,15 +808,58 @@ func TestMapInMapValueSize(t *testing.T) {
 }
 
 func TestIterateEmptyMap(t *testing.T) {
-	hash := createHash()
-	defer hash.Close()
+	makeMap := func(t *testing.T, mapType MapType) *Map {
+		m, err := NewMap(&MapSpec{
+			Type:       mapType,
+			KeySize:    4,
+			ValueSize:  8,
+			MaxEntries: 2,
+		})
+		if errors.Is(err, unix.EINVAL) {
+			t.Skip(mapType, "is not supported")
+		}
+		if err != nil {
+			t.Fatal("Can't create map:", err)
+		}
+		t.Cleanup(func() { m.Close() })
+		return m
+	}
 
-	entries := hash.Iterate()
+	for _, mapType := range []MapType{
+		Hash,
+		SockHash,
+	} {
+		t.Run(mapType.String(), func(t *testing.T) {
+			m := makeMap(t, mapType)
+			entries := m.Iterate()
 
-	var key string
-	var value uint32
-	if entries.Next(&key, &value) != false {
-		t.Error("Empty map should not be iterable")
+			var key string
+			var value uint32
+			if entries.Next(&key, &value) != false {
+				t.Error("Empty hash should not be iterable")
+			}
+			if err := entries.Err(); err != nil {
+				t.Error("Empty hash shouldn't return an error:", err)
+			}
+		})
+	}
+
+	for _, mapType := range []MapType{
+		Array,
+		SockMap,
+	} {
+		t.Run(mapType.String(), func(t *testing.T) {
+			m := makeMap(t, mapType)
+			entries := m.Iterate()
+			var key string
+			var value uint32
+			for entries.Next(&key, &value) {
+				// Some empty arrays like sockmap don't return any keys.
+			}
+			if err := entries.Err(); err != nil {
+				t.Error("Empty array shouldn't return an error:", err)
+			}
+		})
 	}
 }
 
@@ -620,13 +1106,15 @@ func TestMapName(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	if name := internal.CString(info.mapName[:]); name != "test" {
+	if name := internal.CString(info.name[:]); name != "test" {
 		t.Error("Expected name to be test, got", name)
 	}
 }
 
 func TestMapFromFD(t *testing.T) {
 	m := createArray(t)
+	defer m.Close()
+
 	if err := m.Put(uint32(0), uint32(123)); err != nil {
 		t.Fatal(err)
 	}
@@ -638,6 +1126,13 @@ func TestMapFromFD(t *testing.T) {
 		t.Fatal(err)
 	}
 
+	// Both m and m2 refer to the same fd now. Closing either of them will
+	// release the fd to the OS, which then might re-use that fd for another
+	// test. Once we close the second map we might close the re-used fd
+	// inadvertently, leading to spurious test failures.
+	// To avoid this we have to "leak" one of the maps.
+	m2.fd.Forget()
+
 	var val uint32
 	if err := m2.Lookup(uint32(0), &val); err != nil {
 		t.Fatal("Can't look up key:", err)
@@ -761,7 +1256,8 @@ func TestNewMapFromID(t *testing.T) {
 }
 
 func TestMapPinning(t *testing.T) {
-	tmp := tempBPFFS(t)
+	tmp := testutils.TempBPFFS(t)
+	c := qt.New(t)
 
 	spec := &MapSpec{
 		Name:       "test",
@@ -777,11 +1273,18 @@ func TestMapPinning(t *testing.T) {
 		t.Fatal("Can't create map:", err)
 	}
 	defer m1.Close()
+	pinned := m1.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
 
 	if err := m1.Put(uint32(0), uint32(42)); err != nil {
 		t.Fatal("Can't write value:", err)
 	}
 
+	// This is a terrible hack: if loading a pinned map tries to load BTF,
+	// it will get a nil *btf.Spec from this *btf.Map. This is turn will make
+	// btf.NewHandle fail.
+	spec.BTF = new(btf.Map)
+
 	m2, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
 	if err != nil {
 		t.Fatal("Can't create map:", err)
@@ -796,6 +1299,16 @@ func TestMapPinning(t *testing.T) {
 	if value != 42 {
 		t.Fatal("Pinning doesn't use pinned maps")
 	}
+
+	spec.KeySize = 8
+	m3, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+	if err == nil {
+		m3.Close()
+		t.Fatalf("Opening a pinned map with a mismatching spec did not fail")
+	}
+	if !errors.Is(err, ErrMapIncompatible) {
+		t.Fatalf("Opening a pinned map with a mismatching spec failed with the wrong error")
+	}
 }
 
 type benchValue struct {
